(*

  fdo is a command which takes the following subcommands:

  w2fb : converts a file which contains words to a file which
  contains sequences of feature bundles.

  fb2w : converts a file which contains sequences of feature 
  bundles to a file which contains words.

  fbmod : converts a file which contains sequences of feature 
  bundles to a file which contains sequecnes of feature bundles 
  containing only those features specified. Thus, it applies 
  modifies every feature bundle uniformly.

*)

let usage () = print_endline (
    "\nUSAGE: fdo [w2fb|f2wb|modfb|nc] -i inputfile -f featurefile [OPTIONS] \n\n"^
      "See the documentation file for more information.\n");;

let cmds = List.tl (Array.to_list Sys.argv);;
let subcommand = List.hd cmds;;
let cmd_list = List.tl cmds;;

(* the cmd array keeps track of the options. 
   
   0 - contains the inputfile
   1 - contains the delimiter for segments in words 
   2 - contains the delimiter for segments in sets
   3 - contains the featurefile
   4 - contains the delimiter for feature bundles in words
   5 - contains the feature names to be kept in "fbmod" (comma delimited)
   6 - records if there are errors in the command line arguments
   7 - records options for the subcommand "nc" 
   8 - contains the ncmap file
   9 - contains the limiting on number N if computing N partitions by natural class
*)

let a_size = 10;;

let if_pos = 0;;
let xd_pos = 1;;
let xs_pos = 2;;
let ff_pos = 3;;
let fbd_pos = 4;;
let ftn_pos = 5;;
let err_pos = 6;;
let nc_pos = 7;;
let ncm_pos = 8;;
let lim_pos = 9;;

let cmd_array = Array.make a_size "";;
(* Note that the default x delimiter is the empty string *)
Array.set cmd_array xs_pos ",";;  (* default x set delimiter is comma *)
Array.set cmd_array fbd_pos ".";; (* default feature bundle delimiter is period *)


let rec process_args list = 
  match list with
    | [] -> ()
    | "-i"::(inputfile::t) ->
	Array.set cmd_array if_pos inputfile; process_args t
	  
    | "-f"::(featurefile::t) ->
	Array.set cmd_array ff_pos featurefile; process_args t

    | "-xd"::(word_delim::t) ->
	Array.set cmd_array xd_pos word_delim; process_args t

    | "-xs"::(xset_delim::t) ->
	Array.set cmd_array xs_pos xset_delim; process_args t

    | "-fbd"::(fbd_delim::t) ->
	Array.set cmd_array fbd_pos fbd_delim; process_args t

    | "-fn"::(f_names::t) ->
	Array.set cmd_array ftn_pos f_names; process_args t

    | "-s"::t ->
	Array.set cmd_array nc_pos "s"; process_args t

    | "-t"::t ->
	Array.set cmd_array nc_pos "t"; process_args t

    | "-ncm"::(ncmap_file::t) ->
	Array.set cmd_array ncm_pos ncmap_file; process_args t

    | "-lim"::(n::t) ->
	Array.set cmd_array lim_pos n; process_args t

    | _ -> Array.set cmd_array err_pos "true" ;;

process_args cmd_list;;


let xd_delim = Array.get cmd_array xd_pos;; (* delimits segments (i.e. x) in a
					       word (i.e. sequences of x items) *)
let xs_delim = Array.get cmd_array xs_pos;; (* delimits segments (i.e. x) in a
					       set (i.e. sets of x items) *)
let fbd_delim = Array.get cmd_array fbd_pos;; (* delimits feature bundles (i.e. fb) in a
						 sequences (i.e. sequences of fb items) *)


module Word_D = struct let lb = "" let rb = "" let delim = xd_delim end;;
module XSet_D = struct let lb = "" let rb = "" let delim = xs_delim end;;
module FbWord_D = struct let lb = "" let rb = "" let delim = fbd_delim end;;

module PhoneSet = Xset.Make(XSet_D)(Phone);;
module Pi = Xset.Make(Delim.Widebar)(PhoneSet)

module Word = Xlist.Make(Word_D)(Phone);;
module WordSet = Xset.Make(Delim.Newline_no_bs)(Word);;

module F = Feature.Make(Name)(Boolean);;
module Fb = Bundle.Make(Delim.Bar_sqbs)(F);;
module FbWord = Xlist.Make(FbWord_D)(Fb);;
module FbWordSet = Xset.Make(Delim.Newline_no_bs)(FbWord);;

module Fsys = System.Make
  (Phone)(PhoneSet)(Word)(WordSet)(Name)(Boolean)(F)(Fb)(FbWord)
  (FbWordSet);;

module NCMap = Nclass.Make(Phone)(PhoneSet)(Fb)(FbWord)(Fsys);;

module PiSet = Partition.Make(Phone)(PhoneSet)(Pi)(Fsys)(NCMap);;

(* subcommands *)
let w2fb () =
  let featurefile = Array.get cmd_array ff_pos in
  let wordfile = Array.get cmd_array if_pos in
  let ic = if wordfile = "" then stdin else open_in wordfile in
  let t,_,_ = Fsys.of_file featurefile in
  File.fold
    (fun s unit-> 
      let w = Word.of_string s in
      let fbl = Fsys.fbList_of_xList t w in
      let () = FbWord.print fbl in
      unit
    ) ic ();;

let fb2w () =
  let featurefile = Array.get cmd_array ff_pos in
  let fbfile = Array.get cmd_array if_pos in
  let ic = if fbfile = "" then stdin else open_in fbfile in
  let t,phones,_ = Fsys.of_file featurefile in
  File.fold
    (fun s unit-> 
      let fbl = FbWord.of_string s in
      let ws = Fsys.xListSet_of_fbList t fbl (PhoneSet.of_list phones) in
      let () = WordSet.print ws in
      unit
    ) ic ();;

let fbmod () =
  let fbfile = Array.get cmd_array if_pos in
  let fnames = Mstring.to_stringlist "," (Array.get cmd_array ftn_pos) in
  let fb_keep = List.fold_left
    (fun fb fn -> Fb.add 
      (F.make Boolean.max (Name.of_string fn)) fb
    ) Fb.empty fnames 
  in
  let ic = if fbfile = "" then stdin else open_in fbfile in
  File.fold
    (fun s unit-> 
      let fbl = FbWord.of_string s in
      let newfbl = FbWord.map (fun fb -> Fb.modulo_features fb fb_keep) fbl in
      let () = FbWord.print newfbl in
      unit
    ) ic ();;

let nc () = 
  let featurefile = Array.get cmd_array ff_pos in
  let opt = Array.get cmd_array nc_pos in
  let fsys,_,_ = Fsys.of_file featurefile in
  let ncmap = match opt with
      "s" -> NCMap.make_specific fsys
    | "t" -> NCMap.make_total fsys
    | _ -> NCMap.make fsys 
  in
  NCMap.print ncmap;;

let mg () = 
  let featurefile = Array.get cmd_array ff_pos in
  let ncfile = Array.get cmd_array ncm_pos in
  let fsys,_,_ = Fsys.of_file featurefile in
  let ncmap = NCMap.of_file ncfile in
  NCMap.print_mg fsys ncmap;;

let sim () = 
  let featurefile = Array.get cmd_array ff_pos in
  let ncfile = Array.get cmd_array ncm_pos in
  let fsys,_,_ = Fsys.of_file featurefile in
  let ncmap = NCMap.of_file ncfile in
  NCMap.print_scores fsys ncmap;;

let pi () = 
  let featurefile = Array.get cmd_array ff_pos in
  let ncfile = Array.get cmd_array ncm_pos in
  let fsys,_,_ = Fsys.of_file featurefile in
  let ncmap = NCMap.of_file ncfile in
  let n = Array.get cmd_array lim_pos in
  if n = ""
  then PiSet.print (PiSet.make fsys ncmap)
  else PiSet.print (PiSet.make_N (int_of_string n) fsys ncmap)
    

(* let test () =  *)
(*   let featurefile = Array.get cmd_array ff_pos in *)
(*   let fbfile = Array.get cmd_array if_pos in *)
(*   let ic = open_in fbfile in *)
(*   let _ = Fsys.of_file featurefile in *)
(*   File.fold *)
(*     (fun s unit->  *)
(*       let () = print_endline s in *)
(*       unit *)
(*     ) ic ();; *)

let check_args () = 
  let check1 =  
    (subcommand = "w2fb" 
	|| subcommand = "fb2w"
	|| subcommand = "nc") 
    && (Array.get cmd_array ff_pos) = ""
  in
  let check2 = 
    (subcommand = "mg" 
	|| subcommand = "sim")
    && (Array.get cmd_array ff_pos = ""
	|| Array.get cmd_array ncm_pos = "")
  in
  if check1 || check2
  then Array.set cmd_array err_pos "true";;

let main () =
  let () = check_args () in
  if Array.get cmd_array err_pos = "true" 
  then let () = usage () in 
       failwith "Unrecognized arguments. See documentation."
  else 
    match subcommand with
      | "w2fb" -> w2fb ()
      | "fb2w" -> fb2w ()
      | "fbmod" -> fbmod ()
      | "nc" -> nc ()
      | "mg" -> mg ()
      | "sim" -> sim ()
      | "pi" -> pi ()
(*       | "test" -> test ()   *)
      | _ -> let () = usage () in
	     failwith "Unrecognized arguments. See documentation.";;


main ();;
