module type NATURAL_CLASS_TYPE =
sig
  include Xmap.XMAP_TYPE
  type x
  type fsystem
  val make : fsystem -> t
  val make_total : fsystem -> t
  val make_specific : fsystem -> t
  val smallest_containing : t -> x -> x -> key list
  val print_mg : fsystem -> t -> unit
  val sim_score : t -> x -> x -> float
  val print_scores : fsystem -> t -> unit
end


module Make 
  (P: X.X_TYPE)
  (PSet: Xset.XSET_TYPE with type elt = P.t)
  (FB: Xset.XSET_TYPE)
  (FBList: Xlist.XLIST_TYPE with type elt = FB.t)
  (Fsys: System.FEATURE_SYSTEM_TYPE 
  with type x = P.t 
  and type xSet = PSet.t
  and type fb = FB.t
  and type fbList = FBList.t)
  =
  
struct
(* Natural Classes *)

  module NCMap = Xmap.Make(PSet)(FBList)(Delim.Mp)
  include NCMap

  type x = P.t
  type fsystem = Fsys.t

  module FBSet = Set.Make(FB)

  let powerset fb =
    Mset.powerset fb FB.cardinal FB.elements FB.add 
      FB.empty FBSet.add FBSet.empty

  exception No_new_natural_class

  let shorter fb1 fb2 =
    let size1 = FB.cardinal fb1 in
    let size2 = FB.cardinal fb2 in
    if size1 > size2 then fb2
    else fb1

  let longer fb1 fb2 =
    let size1 = FB.cardinal fb1 in
    let size2 = FB.cardinal fb2 in
    if size1 < size2 then fb2
    else fb1

  let admit_all_bundles_for_a_natural_class ncMap ps fb =
    try 
      let fb_list = NCMap.find ps ncMap in 
      if (not (FBList.mem fb fb_list))
      then NCMap.add ps (FBList.build fb fb_list) ncMap
      else ncMap
    with
	Not_found -> 
	  NCMap.add ps (FBList.build fb FBList.empty) ncMap
	  
  let admit_only_shortest_bundle_for_a_natural_class ncMap ps fb =
    try 
      let fb_list = NCMap.find ps ncMap
      in 
      NCMap.add ps 
	(FBList.build (shorter (FBList.hd fb_list) fb) FBList.empty) ncMap
    with
	Not_found -> NCMap.add ps (FBList.build fb FBList.empty) ncMap
	
  let admit_only_longest_bundle_for_a_natural_class ncMap ps fb =
    try 
      let fb_list = NCMap.find ps ncMap
      in 
      NCMap.add ps 
	(FBList.build (longer (FBList.hd fb_list) fb) FBList.empty) ncMap
    with
	Not_found -> NCMap.add ps (FBList.build fb FBList.empty) ncMap


  let find_ncs fsys search_type = 
    let fb = Fsys.feature_value_pairs fsys in
    let _,xlist = Fsys.features_and_xs fsys in
    let total_xset = PSet.of_list xlist in
    let powersetfb = powerset fb in 
    FBSet.fold
      (fun bundle ncmap ->
	let xset = Fsys.xSet_of_fb fsys bundle total_xset in
	search_type ncmap xset bundle
      )
      powersetfb
      NCMap.empty

  let make_total fsys =
    find_ncs fsys admit_all_bundles_for_a_natural_class

  let make fsys =
    find_ncs fsys admit_only_shortest_bundle_for_a_natural_class 

  let make_specific fsys = 
    find_ncs fsys admit_only_longest_bundle_for_a_natural_class 


  (* minimal generalization Albright & Hayes 2002,2003 *)

  let smallest_containing ncmap p1 p2 =
    NCMap.fold 
      (fun k d psl ->
	if (PSet.mem p1 k) && (PSet.mem p2 k)
	then 
	  if List.length psl = 0 
	  then [k]
	  else
	    let size = PSet.cardinal k in
	    let ps_size = PSet.cardinal (List.hd psl) in
	    if size > ps_size then psl
	    else if size < ps_size then [k]
	    else k::psl
	else psl
      )
      ncmap
      []

  let print_mg fsys ncmap = 
    let _,xList = Fsys.features_and_xs fsys in
    let pairList = Mlist.cross xList xList 
    in
    let foldFunction () (p1,p2) =
      let pset = smallest_containing ncmap p1 p2 in 
      print_endline (
	  "("^(P.to_string p1)^","^
	    (P.to_string p2)^")\t"^
	    (Mstring.concat_list " # " (List.map PSet.to_string pset) ""))
    in 
    List.fold_left foldFunction () pairList
      

   (*
     This next part computes similarity of two xs based on 
     shared natural classes
   *)

   let share_xs ps x1 x2 =
     (PSet.mem x1 ps) && (PSet.mem x2 ps)

   let solo_xs ps x1 x2 =
     let mem1 = PSet.mem x1 ps in
     let mem2 = PSet.mem x2 ps in 
     (mem1 && not mem2) || (mem2 && not mem1)

   let sim_score ncMap x1 x2 =
     let foldFunction test k d x =
       if test k x1 x2
       then NCMap.add k d x
       else x
     in 
     let shared_ncMap = 
       NCMap.fold (foldFunction share_xs) ncMap NCMap.empty
     in
     let unshared_ncMap = 
       NCMap.fold (foldFunction solo_xs) ncMap NCMap.empty
     in
     let shared_ncs = NCMap.size shared_ncMap
     in 
     let unshared_ncs = NCMap.size unshared_ncMap
     in 
     (float_of_int shared_ncs) /. 
       ((float_of_int shared_ncs) +. (float_of_int unshared_ncs))

   let print_scores fsys ncMap =
     let _,xList = Fsys.features_and_xs fsys in
     let pairList = Mlist.cross xList xList in
     let foldFunction () (p1,p2) =
       let score = sim_score ncMap p1 p2 in 
       print_string (
	   "("^(P.to_string p1)^","^
	     (P.to_string p2)^")\t"^
	     (string_of_float score)^"\n")
     in 
     List.fold_left foldFunction () pairList


end
