(**

   Nclass implements useful functions related to the natural classes of
   elements in a feature system.

   author: Jeff Heinz

   last updated : October 1, 2006

*)


(** The output signature of {!Nclass.Make}. *)
module type NATURAL_CLASS_TYPE =
sig

  (** The functions of [Xmap.XMAP_TYPE] (see Tools documentation). Type [t] is
      a natural class map. The map is from sets of elements of the space to a
      list of feature bundles (which describe those sets). *)
  include Xmap.XMAP_TYPE

  (** The type of element in the space. *)
  type x

  (** The type of feature system iver the elements in the space. *)
  type fsystem

  (** Returns a natural class map from a feature system. The smallest feature
      bundle is associated with a natural class. *)
  val make : fsystem -> t

  (** Returns a natural class map from a feature system. All feature
      bundles which match some natural class are associate with it. *)
  val make_total : fsystem -> t

  (** Returns a natural class map from a feature system. The most specific
      feature bundle is associated with a natural class. *)
  val make_specific : fsystem -> t

  (** [smallest_containing ncmap x1 x2] returns a list of the natural classes
      with the fewest elements which contain [x1] and [x2]. *)
  val smallest_containing : t -> x -> x -> key list

  (** [print_mg fsystem ncmap] prints to standard output the smallest
      containing natural classes for every pair of elements in [fsystem]. *)
  val print_mg : fsystem -> t -> unit

  (** Returns the natural class similarity metric as described by Frisch,
      Pierrehumbert, et al. *)
  val sim_score : t -> x -> x -> float

  (** [print_scores fsystem ncmap] prints to standard output all the
      similarity scores for every pair of elements in [fsystem]. *)
  val print_scores : fsystem -> t -> unit
end


(** The input signature to {!Nclass.NATURAL_CLASS_TYPE}. *)
module Make 
  (P: X.X_TYPE)
  (PSet: Xset.XSET_TYPE with type elt = P.t)
  (FB: Xset.XSET_TYPE)
  (FBList: Xlist.XLIST_TYPE with type elt = FB.t)
  (Fsys: System.FEATURE_SYSTEM_TYPE 
  with type x = P.t 
  and type xSet = PSet.t
  and type fb = FB.t
  and type fbList = FBList.t)
  :
  NATURAL_CLASS_TYPE
  with type x = P.t
  and type key = PSet.t
  and type data = FBList.t
  and type fsystem = Fsys.t
