module type PARTITIONS_BY_NATURAL_CLASS_TYPE=
sig
  include Xset.XSET_TYPE
  type x
  type block
  type partition
  type fsystem
  type ncmap
  val make_trivial : fsystem -> partition
  val order_partitions : t -> partition list
  val make : fsystem -> ncmap -> t
  val make_N : int -> fsystem -> ncmap -> t
  val get_block : x -> partition -> block
end

module Make
  (P:X.X_TYPE)
  (PS: Xset.XSET_TYPE with type elt = P.t)
  (PSS: Xset.XSET_TYPE with type elt = PS.t)
  (Fsys: System.FEATURE_SYSTEM_TYPE with type x = P.t
				    and type xSet = PS.t)
  (NCMap: Nclass.NATURAL_CLASS_TYPE with type x = P.t
				    and type key = PS.t
				    and type fsystem = Fsys.t)
  =
  
struct
  module PSSS = Xset.Make(Delim.Newline_no_bs)(PSS)
  include PSSS

  type x = P.t
  type block = PS.t
  type partition = PSS.t
  type fsystem = Fsys.t
  type ncmap = NCMap.t

  let make_trivial fsys =
    let _,x_list = 
      Fsys.features_and_xs fsys
    in 
    List.fold_left
      (fun pi elt -> PSS.add (PS.singleton elt) pi)
      PSS.empty
      x_list


  let natural_classes ncmap = 
    let foldFunction key binding set =
      if PS.equal key PS.empty 
      then set
      else PSS.add key set
    in 
    NCMap.fold
      foldFunction
      ncmap
      PSS.empty


  let order_partitions psss =
    let list = PSSS.elements psss
    in 
    let sort_compare x y =
      let size_x = PSS.cardinal x
      in 
      let size_y = PSS.cardinal y
      in 
      if size_x > size_y then 1
      else if size_y > size_x then -1
      else 0
    in 
    List.sort sort_compare list

    
(*   let partitionSet_to_file filename psss =  *)
(*     let opssl = order_partitions psss in *)
(*     File.of_string filename  *)
(*       (Mlist.to_string opssl PSS.to_string "\n" "" "" empty_string) *)

(*   let partitionSet_print ?oc:(oc=stdout) ps =  *)
(*     let l = order_partitions ps in *)
(*     List.iter (fun x -> PSS.print ~oc:oc x) l *)

(*   let partitionSet_of_file = PSSS.of_file *)

  exception NonRefinable
    
  let refine_block_by_nc ncmap block ncset = 
    if ((PS.cardinal block) = 1)
    then raise NonRefinable
    else    
      let rec helper rem_ncs partitions =
	let empty_rem_pss = PSS.is_empty rem_ncs
	in
	let empty_partitions = PSSS.is_empty partitions
	in 
	match (empty_rem_pss, empty_partitions) with
	  | (true,true) -> raise NonRefinable
	  | (true, false) -> partitions
	  | _ ->
	      
	      let ps = PSS.min_elt rem_ncs
	      in 
	      let comp = PS.diff block ps
	      in 
	      if 
		(not (PS.is_empty comp)) &&
		(PS.subset ps block) &&
		(NCMap.mem ps ncmap) && 
		(NCMap.mem comp ncmap)
	      then 
		let xSetSet1 = PSS.singleton comp
		in 
		let xSetSet2 = PSS.add ps xSetSet1
		in 
		let rs = PSS.diff rem_ncs xSetSet2
		in
		helper rs (PSSS.add xSetSet2 partitions)
	      else 
		helper (PSS.remove ps rem_ncs) partitions
      in 
      helper ncset PSSS.empty

  let refine_by_nc ncmap partition ncset = 
    let foldFunction block new_partitions =
      try 
	let subpartitions = refine_block_by_nc ncmap block ncset
	in 
	let remainder = PSS.remove block partition
	in 
	PSSS.fold
	  (fun x -> PSSS.add (PSS.union x remainder))
	  subpartitions
	  new_partitions
      with 
	| NonRefinable -> 
	    new_partitions
    in 
    let new_partitions = 
      PSS.fold 
	foldFunction
	partition    
	PSSS.empty
    in 
    new_partitions


  let make_N n fsys ncMap =
    let all_xs = PS.of_list (snd (Fsys.features_and_xs fsys))
    in 
    let ncset = natural_classes ncMap 
    in
    let rec helper rem_partitions final_partitions = 
      if (PSSS.is_empty rem_partitions)
	|| (PSSS.cardinal final_partitions >= n)
      then 
	final_partitions
      else 
	let new_finals, new_rems =
	  PSSS.fold
	    (fun x (f,r) -> 
	      let new_partitions = refine_by_nc ncMap x ncset
	      in
	      let new_finals = 
		PSSS.union
		  new_partitions
		  f
	      in 
	      let new_rems = 
		PSSS.union
		  new_partitions
		  r
	      in 
	      new_finals,new_rems)
	    rem_partitions
	    (final_partitions,PSSS.empty)
	in 
	helper 
	  new_rems
	  new_finals
    in 
    let top = PSS.singleton all_xs
    in 
    helper (PSSS.singleton top) PSSS.empty;;


  let make fsys ncMap =
    let all_xs = PS.of_list (snd (Fsys.features_and_xs fsys))
    in 
    let ncset = natural_classes ncMap 
    in
    let rec helper rem_partitions final_partitions = 
      if (PSSS.is_empty rem_partitions)
      then 
	final_partitions
      else 
	let new_finals, new_rems =
	  PSSS.fold
	    (fun x (f,r) -> 
	      let new_partitions = refine_by_nc ncMap x ncset
	      in
	      let new_finals = 
		PSSS.union
		  new_partitions
		  f
	      in 
	      let new_rems = 
		PSSS.union
		  new_partitions
		  r
	      in 
	      new_finals,new_rems)
	    rem_partitions
	    (final_partitions,PSSS.empty)
	in 
(* 	let () = print_string  *)
(* 	  ((string_of_int (PSSS.cardinal new_finals))^"\n") *)
(* 	in  *)
	helper 
	  new_rems
	  new_finals
    in 
    let top = PSS.singleton all_xs
    in 
    let partitions = 
      helper 
	(PSSS.singleton top) 
	(PSSS.singleton top)
    in 
(*     let size = PSSS.cardinal partitions  *)
(*     in  *)
(*     let () = print_string ("% There are "^(string_of_int size) *)
(*       ^" ways to partition the segments with natural classes.\n") *)
(*     in  *)
    partitions


  let get_block x partition =
    let pss =
      PSS.filter 
	(fun block -> PS.mem x block)
	partition
    in
    PSS.choose pss



end
