(**

   Partition allows one to compute partitions by natural class.

   author: Jeff Heinz

   last updated : October 1, 2006

*)

(** The output signature of {!Partition.Make}. *)

module type PARTITIONS_BY_NATURAL_CLASS_TYPE=
sig
  (** Standard [Xset.XSET_TYPE] functions (see Tools documentation).  Type [t]
      is a set of partitions by natural class. 
*)
  include Xset.XSET_TYPE

  (** The type of element in the space (e.g. a segment). *)
  type x

  (** A set of elements. *)
  type block

  (** A set of sets of elements. *)
  type partition

  (** The feature system. *)
  type fsystem

  (** The natural class map. *)
  type ncmap

  (** Returns the trivial partition given some feature system. *)
  val make_trivial : fsystem -> partition

  (** Given a set of partitions, orders them by cardinality (the unit
      partition first). *)
  val order_partitions : t -> partition list

(** Computes the partitions by natural class given a feature system and a
    natural class map. *)
  val make : fsystem -> ncmap -> t

  (** [make_N n fsystem ncmap] returns at least the [n] most coarse natural
      classes. *)
  val make_N : int -> fsystem -> ncmap -> t

(** [get_block x pi] returns the set of elements which contain [x] in some
    partition [pi]. *)
  val get_block : x -> partition -> block
end


(** The input signature to {!Partition.PARTITIONS_BY_NATURAL_CLASS_TYPE}. *)

module Make
  (P:X.X_TYPE)
  (PS: Xset.XSET_TYPE with type elt = P.t)
  (PSS: Xset.XSET_TYPE with type elt = PS.t)
  (Fsys: System.FEATURE_SYSTEM_TYPE with type x = P.t
				    and type xSet = PS.t)
  (NCMap: Nclass.NATURAL_CLASS_TYPE with type x = P.t
				    and type key = PS.t
				    and type fsystem = Fsys.t)
  :

  PARTITIONS_BY_NATURAL_CLASS_TYPE
  with type x = P.t
  and type block = PS.t
  and type partition = PSS.t
  and type fsystem = Fsys.t
  and type ncmap = NCMap.t
