module type FEATURE_SYSTEM_TYPE =
sig
  type t
  type x
  type fn
  val to_file : t -> string -> unit
  val of_file : ?col_delim:char -> string -> t * x list * fn list
  type fb
  type xList
  type fbList
  type xSet
  type xListSet
  val fb_of_x : t -> x -> fb
  val fbList_of_xList : t -> xList -> fbList
  val fb_describes_x : t -> fb -> x -> bool 
  val xSet_of_fb : t -> fb -> xSet -> xSet
  val xListSet_of_fbList : t -> fbList -> xSet -> xListSet
  val feature_value_pairs : t -> fb
  val features_and_xs : t -> fn list * x list
end


module Make 
  (P: X.X_TYPE)
  (PSet: Xset.XSET_TYPE with type elt = P.t)
  (PList : Xlist.XLIST_TYPE with type elt = P.t)
  (PListSet : Xset.XSET_TYPE with type elt = PList.t)
  (Name: X.X_TYPE)
  (Value: Value.VALUE_TYPE) 
  (F: Feature.FEATURE_TYPE with type id = Name.t
			  and type value = Value.t)
  (FB: Bundle.BUNDLE_TYPE with type elt = F.t)
  (FBList : Xlist.XLIST_TYPE with type elt = FB.t)

  (FBListSet : Xset.XSET_TYPE with type elt = FBList.t)
  =
  
struct
  module Fmap = Xmap.Make(P)(Value)(Delim.Mp)
  module Fsys = Xmap.Make (Name)(Fmap)(Delim.Mp)

  type t = Fsys.t
  type fn = Name.t
  type x = P.t
  type xSet = PSet.t          
  type xList = PList.t         (* word *)
  type xListSet = PListSet.t   (* wordset *)
  type fb = FB.t               (* feature bundle *) 
  type fbList = FBList.t       (* fb word *)
  type fbListSet = FBListSet.t (* fb wordset *)

  let get_value a f sys =
    try
      Fmap.find a (Fsys.find f sys)
    with Not_found -> raise (Failure("Unspecified"))

  let get_value_string a f sys =
    let aa = P.of_string a in
    let ff = Name.of_string f in
    try
      Value.to_string (get_value aa ff sys)
    with (Failure("Unspecified")) -> "0"

  let of_table t =
    let cols = List.map Name.of_string (Table.get_col_names t) in
    let rows = List.map P.of_string (Table.get_row_names t) in
    let fsys = fst (
	List.fold_left (fun (map,cn) c ->
	  let m = 
	    fst (List.fold_left
		  (fun (m2,rn) r -> 
		    try
		      let data = Value.of_string (Table.get_val_pos rn cn t)
		      in
		      (Fmap.add r data m2), (rn+1)
		    with (Failure("Unspecified")) -> (m2,rn+1)
		  ) (Fmap.empty,1) rows
	    )
	  in
	  ((Fsys.add c m map), cn+1) )
	  (Fsys.empty,1)
	  cols
    )
    in
    fsys, rows, cols
      

  let features_and_xs fsys = 
    let features,xs_map = 
    Fsys.fold
      (fun k fmap (cn,rnm) ->
	k::cn,
	Fmap.fold
	  (fun k d a -> Fmap.add k d a) fmap rnm
	  (* this assumes every phone is specified for some feature; i.e there
	     is no phone which is unspecified everywhere *)
      )
      fsys ([],Fmap.empty)
  in
    let xs = Fmap.fold
      (fun k d a -> k::a)
      xs_map []
    in
    features,xs
      

  let to_table ?col_delim:(col_delim='\t') map =
    let col_names,row_names = features_and_xs map 
    in
    let row_list = List.fold_left
      (fun sl phone -> 
	(List.fold_left
	  (fun l feature -> 
	    let d = try Value.to_string (get_value phone feature map) 
	      with (Failure("Unspecified")) -> "0" 
	    in
	    d::l
	  )
	  [(P.to_string phone)] col_names)::sl
      )
      [] row_names
    in
    let sl = List.map (fun l -> Mstring.concat_list "\t" l "") 
      ((""::(List.map Name.to_string col_names))::(List.map List.rev row_list))
    in
    Table.of_stringlist sl


  let of_file ?col_delim:(col_delim='\t') filename = 
    of_table (Table.of_file ~col_delim:col_delim filename)
    

  let to_file system = Table.to_file (to_table system)


  let feature_value_pairs fsys = 
    Fsys.fold
      (fun fn fm a1 -> 
	Fmap.fold 
	  (fun p v a2 -> FB.add (F.make v fn) a2)
	  fm a1
      ) fsys FB.empty


  (* Bundles to points (x) and back again *)

  let fb_of_x fsys x =
    Fsys.fold 
      (fun f fmap fb -> 
	try let v = Fmap.find x fmap in
	    FB.add (F.make v f) fb
	with Not_found -> fb
      )
      fsys
      FB.empty

  let fbList_of_xList fsys xlist =
    FBList.rev (
	PList.fold_left
	  (fun bl x -> 
	    FBList.build (fb_of_x fsys x) bl
	  )
	  FBList.empty
	  xlist
    )

  let fb_describes_x fsys fb x =
    FB.fold
      (fun f bool ->
	bool && 
	  (try
	      let x_value = get_value x (F.id f) fsys in
	      if x_value = F.value f 
	      then true
	      else false
	    with (Failure("Unspecified")) -> true
	  )  
      ) fb true


  let xSet_of_fb fsys fb xset =
    (*
      This function returns the points in xset which satisfy the
      features in the bundle. It assumes there is no point unspecified eveywhere. 
    *)  
    FB.fold
      (fun f xs -> 
	let v = F.value f in
	let n = F.id f in
	let m = 
	  try Fsys.find n fsys 
	  with Not_found -> raise (Failure("Unrecognized feature"))
	in
	PSet.filter
	  (fun x -> 
	    let xv = 
	      try Fmap.find x m 
	      with Not_found -> raise (Failure("Unrecognized x item"))
	    in  
	    xv = v)
	  xs
      )
      fb
      xset


  let xSet_to_xListSet xset =
    PSet.fold
      (fun x xls -> PListSet.add (PList.build x PList.empty) xls)
      xset PListSet.empty
	
  let xSetList_to_xListSet xsl =
    (* the elements of the returned set are reversed *)
    let xls_hd = xSet_to_xListSet (List.hd xsl) in
    let tail = List.tl xsl in
    List.fold_left 
      (fun xls xs -> 
	let next = xSet_to_xListSet xs in
	Mset.cross PListSet.add 
	  (fun xl1 xl2 -> PList.append xl2 xl1)
	  PListSet.fold PListSet.empty xls next
      ) xls_hd tail


  let xListSet_of_fbList fsys fblist xset = 
    (* this reverses the order as well *)
    let xSetList = 
      FBList.fold_left
	(fun xsl fb -> (xSet_of_fb fsys fb xset)::xsl
	) [] fblist
    in
    xSetList_to_xListSet xSetList





end


module Make_DQ 
  (P: X.X_TYPE)
  (PSet_D: Delim.DELIM_TYPE)
  (PList_D : Delim.DELIM_TYPE)
  (Name: X.X_TYPE)
  (Value: Value.VALUE_TYPE) 
  =

struct 
  module PSet = Xset.Make(PSet_D)(P)
  module PList = Xlist.Make(PList_D)(P)
  module PListSet = Xset.Make(Delim.Newline_no_bs)(PList)

  module F = Feature.Make(Name)(Boolean)
  module Fb = Bundle.Make(Delim.Bar_sqbs)(F)
  module Fbword = Xlist.Make(Delim.Space_no_bs)(Fb)
  module FbwordSet = Xset.Make(Delim.Newline_no_bs)(Fbword)

  module S =
  Make(P)(PSet)(PList)(PListSet)(Name)(Boolean)
    (F)(Fb)(Fbword)(FbwordSet)

  include S

end
(*   let smaller d1 d2 =  *)
(*     if Value.compare d1 d2 >= 0  *)
(*     then d2 *)
(*     else d1 *)

(*   let both m1 m2 =  *)
(*     Fmap.fold  *)
(*       (fun k d1 m ->  *)
(* 	try  *)
(* 	  let d2 = (Fmap.find k m2) in *)
(* 	  Fmap.add k (smaller d1 d2) m *)
(* 	with Not_found -> Fmap.add k d1 m *)
(*       ) *)
(*       m1 *)
(*       Fmap.empty *)

(*   (\* what if there are relations in m2 that are not in m1? *\) *)
      
(*   let minus m =  (\* doesn't work *\) *)
(*     Fmap.fold (fun k d a -> Fmap.add k (Value.opposite d) a) m Fmap.empty *)

