(**

   System implements a feature system and brings together feature bundles, the
   points they describe, the functions which relate the features to the
   elements to the values and so on. Functions allow one to convert between
   elements and feature bundles.

   author: Jeff Heinz

   last updated : October 1, 2006

*)

(** The output signature of {!System.Make} *)

module type FEATURE_SYSTEM_TYPE =
sig
  (** The feature system (maps feature names to maps of elements to values).  *)
  type t

  (** The type of elements in the space which the features describe. *)  
  type x

  (** the type of feature name. *)
  type fn

  (** Writes a feature system to a file. *)
  val to_file : t -> string -> unit

  (** Reads a file where the feature system is given in table form to a table,
      a list of elements, and a list of feature names. *)
  val of_file : ?col_delim:char -> string -> t * x list * fn list

  (** The type for sets of features (i.e. feature bundles). *)
  type fb

  (** The type for sequences of elements (e.g. words) *)
  type xList

  (** The type for sequences of feature bundles. *)
  type fbList

  (** The type for sets of elements. *)
  type xSet

  (** The type for sets of sequences of elements (e.g. sets of words). *)
  type xListSet

  (** Returns a (complete) feature bundle which describes the element. *)
  val fb_of_x : t -> x -> fb

  (** Returns a list of (complete) feature bundles given a list of elements. *)
  val fbList_of_xList : t -> xList -> fbList

  (** [fb_describes x system fb x] returns [true] iff the value-feature name
      pairs in [fb] are true of [x]. *)
  val fb_describes_x : t -> fb -> x -> bool 

  (** [xSet_of_fb system fb xs] returns a subset of [xs] such that every
      element in the subset is described by [fb]. *)
  val xSet_of_fb : t -> fb -> xSet -> xSet

  (** [xListSet_of_fbList system fbList xSet] returns a set of sequences of
      elements in [xSet] (e.g. a set of words) which match [fbList] *)
  val xListSet_of_fbList : t -> fbList -> xSet -> xListSet

  (** Returns all feature-value pairs in a feature system. *)
  val feature_value_pairs : t -> fb

  (** Returns the all the feature names and elements in the feature system. *)
  val features_and_xs : t -> fn list * x list
end



(** The module below takes a number of modules to produce a module of
    FEATURE_SYSTEM_TYPE. *)

module Make 
  (P : X.X_TYPE)
  (PSet: Xset.XSET_TYPE with type elt = P.t)
  (PList : Xlist.XLIST_TYPE with type elt = P.t)
  (PListSet : Xset.XSET_TYPE with type elt = PList.t)
  (Name : X.X_TYPE)
  (Value: Value.VALUE_TYPE) 
  (F : Feature.FEATURE_TYPE with type id = Name.t
			  and type value = Value.t)
  (FB : Bundle.BUNDLE_TYPE with type elt = F.t)
  (FBList : Xlist.XLIST_TYPE with type elt = FB.t)
  (FBListSet : Xset.XSET_TYPE with type elt = FBList.t)

  : FEATURE_SYSTEM_TYPE
  with type x = P.t
  and type fb = FB.t
  and type xList = PList.t
  and type fbList = FBList.t
  and type xSet = PSet.t
  and type xListSet = PListSet.t


(** The module below only takes fewer modules and returns a module of
    FEATURE_SYSTEM_TYPE. *)
module Make_DQ 
  (P: X.X_TYPE)
  (PSet_D: Delim.DELIM_TYPE)
  (PList_D : Delim.DELIM_TYPE)
  (Name: X.X_TYPE)
  (Value: Value.VALUE_TYPE) 

  : FEATURE_SYSTEM_TYPE

